<?php
declare(strict_types=1);

namespace Inti\Media\Plugin;

use Inti\Media\Exception\InvalidArgumentException;
use Inti\Media\Model\ImageSize;
use Inti\Wp\Plugin\PluginInterface;
use Laminas\Hydrator\HydratorInterface;
use Laminas\Hydrator\ReflectionHydrator;
use League\Flysystem\FilesystemException;
use League\Flysystem\FilesystemOperator;
use Symfony\Contracts\Cache\CacheInterface;
use function add_filter;
use function boolval;
use function get_option;
use function imagecreatefromstring;
use function imagesx;
use function imagesy;
use function in_array;
use function intval;
use function is_string;
use function md5;
use function preg_match;
use function sprintf;

/**
 * Class GathImage
 *
 * @package Inti\Media\Plugin
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/oktan-nautika/inti-media
 */
final class GathImage implements PluginInterface
{
    private CacheInterface $cache;
    private FilesystemOperator $filesystem;
    private ?HydratorInterface $hydrator = null;

    /** @var string[] */
    private array $wpSizes = [
        'large',
        'medium',
        'thumb',
        'thumbnail',
    ];

    public function __construct(CacheInterface $cache, FilesystemOperator $filesystem)
    {
        $this->cache = $cache;
        $this->filesystem = $filesystem;
    }

    public function attachEarly(): void
    { }

    public function attach(): void
    {
        add_filter('wp_get_attachment_image_src', [$this, 'get_attachment_image_src'], 10, 3);
    }

    /**
     * @param array|false $image
     * @param string|int $attachment_id
     * @param string|int[] $size
     * @return array|false
     * @throws FilesystemException
     * @throws \Psr\Cache\InvalidArgumentException
     */
    public function get_attachment_image_src($image, $attachment_id, $size)
    {
        if ($image === false && is_string($attachment_id) && preg_match('/^gath-(.*)$/', $attachment_id, $matches)) {
            $imagePath = $matches[1];

            if (preg_match('/(.*)\.([a-z0-9]{32})(\.[^.]+)$/', $imagePath, $matches)) {
                $imagePath = $matches[1] . $matches[3];
                $hash = $matches[2];
            }

            if (!$this->filesystem->fileExists('image://' . $imagePath)) {
                return $image;
            }

            $requestedSize = $this->getRequestedImageSize($size);

            if ($requestedSize instanceof $imagePath) {
                $imagePath = $this->getResizedImagePath($imagePath, $requestedSize);
                $resized = true;
            }

            $imageUrl = sprintf('/media/%s%s', $imagePath, isset($hash) ? '?' . $hash : '');
            $realSize = $this->getRealImageSize($imagePath);

            return [$imageUrl, ...$realSize, $resized ?? false];
        }

        return $image;
    }

    /**
     * @param string $path
     * @return array
     * @throws \Psr\Cache\InvalidArgumentException
     */
    private function getRealImageSize(string $path): array
    {
        $key = md5($path);

        $meta = $this->cache->get($key, function () use ($path) {
            $image = imagecreatefromstring($this->filesystem->read('image://' . $path));

            return [
                'width' => imagesx($image),
                'height' => imagesy($image),
            ];
        });

        return [$meta['width'], $meta['height']];
    }

    /**
     * @param string $imagePath
     * @param ImageSize $size
     * @return string
     */
    private function getResizedImagePath(string $imagePath, ImageSize $size): string
    {
        return sprintf(
            '%s%dx%d/%s',
            $size->isCropped() ? 'thumb/' : '',
            $size->getWidth(),
            $size->getHeight(),
            $imagePath,
        );
    }

    /**
     * @param string|int[] $requestedSize
     * @return ImageSize|null
     */
    private function getRequestedImageSize($requestedSize): ?ImageSize
    {
        if (!is_string($requestedSize) && !is_array($requestedSize)) {
            throw InvalidArgumentException::typeForArgument('$requestedSize', 'string or array', $requestedSize);
        }

        if (is_array($requestedSize)) {
            return $this->hydrateImageSize(...$requestedSize);
        }

        if ('full' === $requestedSize) {
            return null;
        }

        if (in_array($requestedSize, $this->wpSizes)) {
            return $this->getWpImageSize($requestedSize);
        }

        return $this->getAddedImageSize($requestedSize);
    }

    private function getWpImageSize(string $size): ImageSize
    {
        if (!in_array($size, $this->wpSizes)) {
            throw InvalidArgumentException::scalarValueForArgument('$size', $this->wpSizes, $size);
        }

        if ($size === 'thumb') {
            $size = 'thumbnail';
        }

        return $this->hydrateImageSize(
            intval(get_option("{$size}_size_w", null)),
            intval(get_option("{$size}_size_h", null)),
            boolval(get_option("{$size}_crop", false))
        );
    }

    private function getAddedImageSize(string $size): ImageSize
    {
        global $_wp_additional_image_sizes;

        if (!array_key_exists($size, $_wp_additional_image_sizes)) {
            throw InvalidArgumentException::scalarValueForArgument(
                '$size',
                array_keys($_wp_additional_image_sizes),
                $size
            );
        }

        return $this->hydrateImageSize(
            $_wp_additional_image_sizes[$size]['width'] ?? null,
            $_wp_additional_image_sizes[$size]['height'] ?? null,
            boolval($_wp_additional_image_sizes[$size]['crop'] ?? false)
        );
    }

    private function hydrateImageSize(?int $width, ?int $height, bool $cropped = false): ImageSize
    {
        if ($this->hydrator === null) {
            $this->hydrator = new ReflectionHydrator();
        }

        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->hydrator->hydrate([
            'width' => $width,
            'height' => $height,
            'cropped' => $cropped,
        ], new ImageSize());
    }
}
